/* Copyright (C) 2013-2016 Vincent Forest (vaplv@free.fr)
 *
 * The RSys library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The RSys library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the RSys library. If not, see <http://www.gnu.org/licenses/>. */

/*
 * Internal header used to generate funcs on float vector of X dimensions
 */
#if !defined(FLOATX_DIMENSION__)
  #error Missing arguments
#endif

#if defined(FLOATX_FUNC__)
  #error Unexpected macro definition
#endif

#include "math.h"
#include "rsys.h"
#include <math.h>

#ifdef COMPILER_GCC
  #pragma GCC push_options
  #pragma GCC optimize("unroll-loops")
#endif

STATIC_ASSERT(FLOATX_DIMENSION__ > 1, Unexpected_value);

#define FLOATX_FUNC__(Func) \
  CONCAT(CONCAT(CONCAT(f, FLOATX_DIMENSION__), _), Func)

/* Helper macro */
#define SIZEOF_FLOATX__ sizeof(float[FLOATX_DIMENSION__])

#if FLOATX_DIMENSION__ <= 4
static FINLINE float*
CONCAT(f, FLOATX_DIMENSION__)
  (float* dst
  ,const float x
  ,const float y
#if FLOATX_DIMENSION__ > 2
  ,const float z
#endif
#if FLOATX_DIMENSION__ > 3
  ,const float w
#endif
  )
{
  ASSERT(dst);
  dst[0] = x;
  dst[1] = y;
#if FLOATX_DIMENSION__ > 2
  dst[2] = z;
#endif
#if FLOATX_DIMENSION__ > 3
  dst[3] = w;
#endif
  return dst;
}
#endif

static FINLINE float*
FLOATX_FUNC__(splat)(float* dst, const float val)
{
  int i;
  ASSERT(dst);
  FOR_EACH(i, 0, FLOATX_DIMENSION__)
    dst[i] = val;
  return dst;
}

static FINLINE float*
FLOATX_FUNC__(set__)(float* dst, const float* src)
{
  int i;
  ASSERT(dst && src);
  ASSERT(!MEM_AREA_OVERLAP(dst, SIZEOF_FLOATX__, src, SIZEOF_FLOATX__));
  FOR_EACH(i, 0, FLOATX_DIMENSION__)
    dst[i] = src[i];
  return dst;
}

static FINLINE float*
FLOATX_FUNC__(set)(float* dst, const float* src)
{
  ASSERT(dst && src);
  if(!MEM_AREA_OVERLAP(dst, SIZEOF_FLOATX__, src, SIZEOF_FLOATX__)) {
    return FLOATX_FUNC__(set__)(dst, src);
  } else {
    float tmp[FLOATX_DIMENSION__];
    return FLOATX_FUNC__(set__)(dst, FLOATX_FUNC__(set__)(tmp, src));
  }
}

static FINLINE float
FLOATX_FUNC__(dot)(const float* a, const float* b)
{
  float dot = 0.0f;
  int i;
  ASSERT(a && b);
  FOR_EACH(i, 0, FLOATX_DIMENSION__)
    dot += a[i] * b[i];
  return dot;
}

static FINLINE float
FLOATX_FUNC__(len)(const float* a)
{
  ASSERT(a);
  return (float)sqrt((double)FLOATX_FUNC__(dot)(a, a));
}

static FINLINE float
FLOATX_FUNC__(normalize)(float* dst, const float* a)
{
  float tmp[FLOATX_DIMENSION__];
  float len, rcp_len;
  int i;
  ASSERT(dst && a);

  len = FLOATX_FUNC__(len)(a);
  if(len == 0.f)
    return len;

  rcp_len = 1.0f / len;
  FOR_EACH(i, 0, FLOATX_DIMENSION__)
    tmp[i] = a[i] * rcp_len;
  FLOATX_FUNC__(set__)(dst, tmp);
  return len;
}

static FINLINE char
FLOATX_FUNC__(is_normalized)(const float* a)
{
  return eq_epsf(FLOATX_FUNC__(len)(a), 1.f, 1.e-6f);
}

static FINLINE float*
FLOATX_FUNC__(add)(float* dst, const float* a, const float* b)
{
  float tmp[FLOATX_DIMENSION__];
  int i;
  ASSERT(dst && a && b);
  FOR_EACH(i, 0, FLOATX_DIMENSION__)
    tmp[i] = a[i] + b[i];
  return FLOATX_FUNC__(set__)(dst, tmp);
}

static FINLINE float*
FLOATX_FUNC__(addf)(float* dst, const float* a, const float f)
{
  float tmp[FLOATX_DIMENSION__];
  int i;
  ASSERT(dst && a);
  FOR_EACH(i, 0, FLOATX_DIMENSION__)
    tmp[i] = a[i] + f;
  return FLOATX_FUNC__(set__)(dst, tmp);
}

static FINLINE float*
FLOATX_FUNC__(sub)(float* dst, const float* a, const float* b)
{
  float tmp[FLOATX_DIMENSION__];
  int i;
  ASSERT(dst && a && b);
  FOR_EACH(i, 0, FLOATX_DIMENSION__)
    tmp[i] = a[i] - b[i];
  return FLOATX_FUNC__(set__)(dst, tmp);
}

static FINLINE float*
FLOATX_FUNC__(subf)(float* dst, const float* a, const float f)
{
  float tmp[FLOATX_DIMENSION__];
  int i;
  ASSERT(dst && a);
  FOR_EACH(i, 0, FLOATX_DIMENSION__)
    tmp[i] = a[i] - f;
  return FLOATX_FUNC__(set__)(dst, tmp);
}

static FINLINE float*
FLOATX_FUNC__(mul)(float* dst, const float* a, const float* b)
{
  float tmp[FLOATX_DIMENSION__];
  int i;
  ASSERT(dst && a && b);
  FOR_EACH(i, 0, FLOATX_DIMENSION__)
    tmp[i] = a[i] * b[i];
  return FLOATX_FUNC__(set__)(dst, tmp);
}

static FINLINE float*
FLOATX_FUNC__(mulf)(float* dst, const float* a, const float f)
{
  float tmp[FLOATX_DIMENSION__];
  int i;
  ASSERT(dst && a);
  FOR_EACH(i, 0, FLOATX_DIMENSION__)
    tmp[i] = a[i] * f;
  return FLOATX_FUNC__(set__)(dst, tmp);
}

static FINLINE float*
FLOATX_FUNC__(div)(float* dst, const float* a, const float* b)
{
  float tmp[FLOATX_DIMENSION__];
  int i;
  ASSERT(dst && a && b);
  FOR_EACH(i, 0, FLOATX_DIMENSION__)
    tmp[i] = a[i] / b[i];
  return FLOATX_FUNC__(set__)(dst, tmp);
}

static FINLINE float*
FLOATX_FUNC__(divf)(float* dst, const float* a, const float f)
{
  float tmp[FLOATX_DIMENSION__];
  int i;
  ASSERT(dst && a);
  FOR_EACH(i, 0, FLOATX_DIMENSION__)
    tmp[i] = a[i] / f;
  return FLOATX_FUNC__(set__)(dst, tmp);
}

static FINLINE float*
FLOATX_FUNC__(minus)(float* dst, const float* a)
{
  float tmp[FLOATX_DIMENSION__];
  int i;
  ASSERT(dst && a);
  FOR_EACH(i, 0, FLOATX_DIMENSION__)
    tmp[i] = -a[i];
  return FLOATX_FUNC__(set__)(dst, tmp);
}

static FINLINE float
FLOATX_FUNC__(sum)(const float* a)
{
  float f = 0.0f;
  int i = 0;
  ASSERT(a);
  do {
    f += a[i];
    ++i;
  } while(i < FLOATX_DIMENSION__);
  return f;
}

static FINLINE float*
FLOATX_FUNC__(lerp)
  (float* dst,
   const float* from,
   const float* to,
   const float t)
{
  ASSERT(dst && from && to);
  if(t <= 0.f) {
    FLOATX_FUNC__(set)(dst, from);
  } else if(t >= 1.0f) {
    FLOATX_FUNC__(set)(dst, to);
  } else {
    float tmp[FLOATX_DIMENSION__];
    int i;
    FOR_EACH(i, 0, FLOATX_DIMENSION__)
      tmp[i] = from[i] + t * (to[i] - from[i]);
    FLOATX_FUNC__(set__)(dst, tmp);
  }
  return dst;
}

static FINLINE char
FLOATX_FUNC__(eq)(const float* a, const float* b)
{
  int i = 0;
  char is_eq = 1;
  ASSERT(a && b);
  do {
    is_eq = a[i] == b[i];
    ++i;
  } while(i < FLOATX_DIMENSION__  && is_eq);
  return is_eq;
}

static FINLINE char
FLOATX_FUNC__(eq_eps)(const float* a, const float* b, const float eps)
{
  int i = 0;
  char is_eq = 1;
  ASSERT(a && b);
  do {
    is_eq = eq_epsf(a[i], b[i], eps);
    ++i;
  } while(i < FLOATX_DIMENSION__ && is_eq);
  return is_eq;
}

static FINLINE float*
FLOATX_FUNC__(max)(float* dst, const float* a, const float* b)
{
  float tmp[FLOATX_DIMENSION__];
  int i;
  ASSERT(dst && a && b);
  FOR_EACH(i, 0, FLOATX_DIMENSION__)
    tmp[i] = a[i] > b[i] ? a[i] : b[i];
  return FLOATX_FUNC__(set__)(dst, tmp);
}

static FINLINE float*
FLOATX_FUNC__(min)(float* dst, const float* a, const float* b)
{
  float tmp[FLOATX_DIMENSION__];
  int i;
  ASSERT(dst && a && b);
  FOR_EACH(i, 0, FLOATX_DIMENSION__)
    tmp[i] = a[i] < b[i] ? a[i] : b[i];
  return FLOATX_FUNC__(set__)(dst, tmp);
}

#undef FLOATX_DIMENSION__
#undef FLOATX_FUNC__

#ifdef COMPILER_GCC
  #pragma GCC pop_options
#endif

